searchNodes=[{"ref":"Lab42.StateMachine.html","title":"Lab42.StateMachine","type":"module","doc":"Lab42.StateMachine Synopsis A simple State Machine operating on a list of input values, a map of transition definitions and an accumulator. What is the Transition Definitions Map? It maps each state to a list of transitions. Each transition is of the following format, as defined by type transition_t {trigger, transformer, updater, new_state} State, did you say state? Yes I did, and what I mean by it is that the State Machine keeps track of its state (very surprising) and some data, passed into by the user. How do these Transitions transform the input? The State Machine processes its input in a loop in which, depending on the current input element, designated as input and the current state, designated as current_state a transition is triggered. Data is passed in like an accumulator in Enum.reduce it is designated as data. The triggered transition will define the current_state of the next loop, if any, and perform actions defining what goes to the output and how data changes. As one might guess these actions are performed by the transformer and the updater of the triggered transition. So what is the contract? Starting the whole thing. Create the Transition Defintions Map, knowing that the State Machine will start with current_state equal to :start, yes I know, naming is difficult ;). Then run the whole thing like that: Lab42.StateMachine.run(input, my_data, %{}) The empty map passed in will cause havoc though iex(0)&gt; input = ~w(alpha) ...(0)&gt; my_data = %{n: 42} ...(0)&gt; Lab42.StateMachine.run(input, my_data, %{}) {:error, &quot;No transitions found for current state&quot;, :start} Specifying Transitions A minimal example might be a line counter iex(1)&gt; input = ~w{alpha beta} ...(1)&gt; count = 0 ...(1)&gt; states = %{ ...(1)&gt; start: [ {true, fn {_, line} -&gt; line end, fn count, _ -&gt; count + 1 end, :start} ] ...(1)&gt; } ...(1)&gt; run(input, count, states) {:start, ~w{alpha beta}, 2} N.B. That the true trigger always matches, other triggers are either regular expressions, which are used as follows: Regex.match(trigger, input), or functions, which are used as one might easily guess, thusly: trigger.(input) One could argue that a default behavior of copying the current input to the output might be convenient, but that might lead to difficulties in debugging state machines. (Maybe in later versions with an option for run?) On the same token some StateMachines, like the counter above collect the output without needing it, although we will learn below how to avoid this, a global option to not collect will make the Transitions Map more concise. Therefore the following will happen iex(2)&gt; input = ~w{alpha beta} ...(2)&gt; states = %{ ...(2)&gt; start: [] ...(2)&gt; } ...(2)&gt; run(input, nil, states) {:error, &quot;No trigger matched the current input \\&quot;alpha\\&quot;&quot;, :start} Let us return to the correctly working example, let us simplify that rather expressive transition iex(3)&gt; input = ~w{alpha beta} ...(3)&gt; states = %{ ...(3)&gt; start: [ {true, :id, fn count, _ -&gt; count + 1 end, :start} ] ...(3)&gt; } ...(3)&gt; run(input, 0, states) {:start, ~w(alpha beta), 2} So we can use a shortcut for copying the input to the output, that is better already, but still, why create the output that is not needed, let us use the atom form of the transformer function iex(4)&gt; input = ~w{alpha beta} ...(4)&gt; states = %{ ...(4)&gt; start: [ {true, fn _ -&gt; :ignore end, fn count, _ -&gt; count + 1 end, :start} ] ...(4)&gt; } ...(4)&gt; run(input, 0, states) {:start, ~w(), 2} As there is a shortcut for :id so there is one for :ignore iex(5)&gt; input = ~w{alpha beta} ...(5)&gt; states = %{ ...(5)&gt; start: [ {true, :ignore, fn count, _ -&gt; count + 1 end, :start} ] ...(5)&gt; } ...(5)&gt; run(input, 0, states) {:start, ~w(), 2} But what if we want to have :ignore in the output? Let us assume that we want to replace all &quot;alphas&quot; with :ignore. We can use the tuple form of the transformer in this case. This example also demonstrates that we do not need to specify the updater and the new state if these do not change and that even pushing the input to the output can be omitted, therefore the transitions {true, :id} and {true} have the same semantics. And a third simplifaction is that we can omit to pass nil as data, but note that it will present in the result. iex(6)&gt; input = ~w{alpha beta alpha} ...(6)&gt; states = %{ ...(6)&gt; start: [ ...(6)&gt; {~r{alpha}, fn _ -&gt; {:push, :ignore} end}, ...(6)&gt; {true}] } ...(6)&gt; run(input, states) {:start, [:ignore, &quot;beta&quot;, :ignore], nil} Now let us look at a complex example, that will use the following features, not yet explored: Stopping with the :halt atom form of the transformer, by its shortcut Stopping with the {:halt, value} tuple form of the transformer Creating a function to rerun with the same states Using constant functions Using function triggers iex(7)&gt; summer = fn data, {_, input} -&gt; %{data|sum: data.sum + input} end ...(7)&gt; states = %{ ...(7)&gt; start: [ ...(7)&gt; {&amp;(&amp;1&gt;8), :halt}, ...(7)&gt; {&amp;(&amp;1&lt;1), fn {_, input} -&gt; {:halt, input} end, constant(:negative)}, ...(7)&gt; {&amp;(rem(&amp;1,2)==0), push_constant(:even), summer, :even}, ...(7)&gt; {true} ], ...(7)&gt; even: [ ...(7)&gt; {&amp;(rem(&amp;1,2)==0), :halt, constant(:more_evens)}, ...(7)&gt; {true, :id, summer } ...(7)&gt; ] } ...(7)&gt; state_machine = make_machine(%{sum: 0}, states) ...(7)&gt; [ ...(7)&gt; state_machine.(1..10|&gt;Enum.into([])), ...(7)&gt; state_machine.([1, 2, 3]), ...(7)&gt; state_machine.([1, 9]), ...(7)&gt; state_machine.([1, -1]), ...(7)&gt; state_machine.([1, 3, 5]) ] [ {:even, [1, :even, 3], :more_evens}, {:even, [1, :even, 3], %{sum: 5}}, {:start, [1], %{sum: 0}}, {:start, [1, -1], :negative}, {:start, [1, 3, 5], %{sum: 0}} ] Summary of Shortcuts and Transformer Semantics As we have seen in the examples above, the StateMachine is driven by the Transitions Map and the result of the transformer function. The returned data value is influenced by the result of the updater function. In order to write shorter Transition Maps transitions can be shortened and symbolic function names can be used. Let us see how transitions are normalized by the StateMachine before they are executed. SpecifiedNormalizeRemarks {trigger, function, function, state}same valuealready normalized {trigger, function, function}{trigger, function, function, current_state}normalization is done during tuntime, thusly current_state is known {trigger, function }{trigger, function, identity2, current_state}given the interface of updater identity2 is defined as fn d, _ -&gt; d end |{trigger} | {trigger, identity1, identity2, current_state} | identity1 is for the transformer and thusly fn {_, v} -&gt; v end As can be seen sometimes the defaults, are not available, because we need to provide the new_state in the transition. In that case the following symbols can be used for the transformer Symbolic TransformerExpanded to :idfn {_, v} -&gt; v end :ignorefn _ -&gt; :ignore end :haltfn _ -&gt; :halt end For the updater there is only :id which expands to fn d, _ -&gt; d end as mentioned above. In action that would give iex(8)&gt; states = %{ ...(8)&gt; start: [{true, :id, :id, :stop}], ...(8)&gt; stop: [{true, fn _ -&gt; {:halt, &quot;...&quot;} end, :id}] ...(8)&gt; } ...(8)&gt; run(~w[Headline Rest MoreRest], states) {:stop, ~w[Headline ...], nil}"},{"ref":"Lab42.StateMachine.html#constant/1","title":"Lab42.StateMachine.constant/1","type":"function","doc":"A helper creating an updater function that sets data to a constant value"},{"ref":"Lab42.StateMachine.html#make_machine/2","title":"Lab42.StateMachine.make_machine/2","type":"function","doc":"A convenience helper to run the same state machine on different inputs"},{"ref":"Lab42.StateMachine.html#push_constant/1","title":"Lab42.StateMachine.push_constant/1","type":"function","doc":"A helper creating a transformer function that pushes a constant value to the output"},{"ref":"Lab42.StateMachine.html#run/3","title":"Lab42.StateMachine.run/3","type":"function","doc":""},{"ref":"Lab42.StateMachine.html#t:empty_t/0","title":"Lab42.StateMachine.empty_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:error_t/0","title":"Lab42.StateMachine.error_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:incomplete_transition_t/0","title":"Lab42.StateMachine.incomplete_transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:match_t/0","title":"Lab42.StateMachine.match_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:maybe/1","title":"Lab42.StateMachine.maybe/1","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:normalized_transitions_t/0","title":"Lab42.StateMachine.normalized_transitions_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:ok_t/0","title":"Lab42.StateMachine.ok_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:result_t/0","title":"Lab42.StateMachine.result_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:state_t/0","title":"Lab42.StateMachine.state_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:symbolic_transformer_t/0","title":"Lab42.StateMachine.symbolic_transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:symbolic_updater_t/0","title":"Lab42.StateMachine.symbolic_updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transformer_t/0","title":"Lab42.StateMachine.transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transition_map_t/0","title":"Lab42.StateMachine.transition_map_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:transition_t/0","title":"Lab42.StateMachine.transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:trigger_t/0","title":"Lab42.StateMachine.trigger_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.html#t:updater_t/0","title":"Lab42.StateMachine.updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html","title":"Lab42.StateMachine.Runner","type":"module","doc":"Runs the state machine by finding and executing transitions"},{"ref":"Lab42.StateMachine.Runner.html#halt_transfo/1","title":"Lab42.StateMachine.Runner.halt_transfo/1","type":"function","doc":"Convenience transformer function to stop the state machine, can be used with the atom :halt"},{"ref":"Lab42.StateMachine.Runner.html#ident_transfo/1","title":"Lab42.StateMachine.Runner.ident_transfo/1","type":"function","doc":"Convenience function to push an input to the output without changing it, can be used with the atom :id in the transformer position of a transition, e.g. {~r{alpha}, :id, fn count, _ -&gt; count + 1 end}"},{"ref":"Lab42.StateMachine.Runner.html#ident_updater/2","title":"Lab42.StateMachine.Runner.ident_updater/2","type":"function","doc":"Convenience function to not change the data. It can be used with the atom :id in the updater position of a transition, e.g. {~r{alpha}, fn {_, line} -&gt; String.reverse(line), :id}"},{"ref":"Lab42.StateMachine.Runner.html#ignore_input/1","title":"Lab42.StateMachine.Runner.ignore_input/1","type":"function","doc":"Convenience function to ignore an input, it can be used with the atom :ignore in the transformer position of a transition, e.g. {~r{alpha}, :ignore, fn count, _ -&gt; count + 1 end}"},{"ref":"Lab42.StateMachine.Runner.html#t:empty_t/0","title":"Lab42.StateMachine.Runner.empty_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:error_t/0","title":"Lab42.StateMachine.Runner.error_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:incomplete_transition_t/0","title":"Lab42.StateMachine.Runner.incomplete_transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:match_t/0","title":"Lab42.StateMachine.Runner.match_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:maybe/1","title":"Lab42.StateMachine.Runner.maybe/1","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:normalized_transitions_t/0","title":"Lab42.StateMachine.Runner.normalized_transitions_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:ok_t/0","title":"Lab42.StateMachine.Runner.ok_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:result_t/0","title":"Lab42.StateMachine.Runner.result_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:state_t/0","title":"Lab42.StateMachine.Runner.state_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:symbolic_transformer_t/0","title":"Lab42.StateMachine.Runner.symbolic_transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:symbolic_updater_t/0","title":"Lab42.StateMachine.Runner.symbolic_updater_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transformer_t/0","title":"Lab42.StateMachine.Runner.transformer_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transition_map_t/0","title":"Lab42.StateMachine.Runner.transition_map_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:transition_t/0","title":"Lab42.StateMachine.Runner.transition_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:trigger_t/0","title":"Lab42.StateMachine.Runner.trigger_t/0","type":"type","doc":""},{"ref":"Lab42.StateMachine.Runner.html#t:updater_t/0","title":"Lab42.StateMachine.Runner.updater_t/0","type":"type","doc":""}]